#ifndef _QUAKE_RESOURCE_H
#define _QUAKE_RESOURCE_H

//{{NO_DEPENDENCIES}}
// Microsoft Developer Studio generated include file.
// Used by winquake.rc
//
#define IDS_STRING1	 1
#define IDI_ICON2	 1
#define IDD_DIALOG1	 108
#define IDD_PROGRESS 109
#define IDC_PROGRESS 1000

// Next default values for new objects
//
#ifdef APSTUDIO_INVOKED
#ifndef APSTUDIO_READONLY_SYMBOLS
#define _APS_NEXT_RESOURCE_VALUE 113
#define _APS_NEXT_COMMAND_VALUE	 40001
#define _APS_NEXT_CONTROL_VALUE	 1004
#define _APS_NEXT_SYMED_VALUE	 101
#endif
#endif

#endif /* _QUAKE_RESOURCE_H */
