/* module tracker decoding support using libxmp */
#if !defined(_SND_XMP_H_)
#define _SND_XMP_H_

#if defined(USE_CODEC_XMP)

extern snd_codec_t xmp_codec;

#endif /* USE_CODEC_XMP */

#endif /* ! _SND_XMP_H_ */
