#include <winresrc.h>
#include "../Quake/quakever.h"

icon ICON "vkQuake.ico"

////////////////////////////////////////////////////////////////

#define VER_FILEVERSION				VKQUAKE_VERSION_MAJOR,VKQUAKE_VERSION_MINOR,VKQUAKE_VER_PATCH,0
#define VER_FILEVERSION_STR			QS_STRINGIFY(VKQUAKE_VERSION) "." QS_STRINGIFY(VKQUAKE_VER_PATCH) VKQUAKE_VER_SUFFIX

#define VER_PRODUCTVERSION			VER_FILEVERSION
#define VER_PRODUCTVERSION_STR		VER_FILEVERSION_STR

#ifdef DEBUG
	#define VER_DEBUG				VS_FF_DEBUG
#else
	#define VER_DEBUG				0
#endif

//////////////////////////////////////////////////////////////// 

VS_VERSION_INFO VERSIONINFO
FILEVERSION		VER_FILEVERSION
PRODUCTVERSION	VER_PRODUCTVERSION
FILEFLAGSMASK	VS_FFI_FILEFLAGSMASK
FILEFLAGS		VER_DEBUG
FILEOS			VOS__WINDOWS32
FILETYPE		VFT_APP
FILESUBTYPE		0
BEGIN
	BLOCK "StringFileInfo"
	BEGIN
		BLOCK "040904E4"
		BEGIN
			VALUE "FileDescription",	"vkQuake Quake engine"
			VALUE "FileVersion",		VER_FILEVERSION_STR
			VALUE "LegalCopyright",		"\xA9 2024"
			VALUE "OriginalFilename",	"vkQuake.exe"
			VALUE "ProductName",		"vkQuake"
			VALUE "ProductVersion",		VER_PRODUCTVERSION_STR
		END
	END

	BLOCK "VarFileInfo"
	BEGIN
		VALUE "Translation", 0x409, 1252
	END
END
