/* Ogg/Opus streaming music support. */

#if !defined(_SND_OPUS_H_)
#define _SND_OPUS_H_ 1

#if defined(USE_CODEC_OPUS)

extern snd_codec_t opus_codec;

#endif /* USE_CODEC_OPUS */

#endif /* ! _SND_OPUS_H_ */
