/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others
Copyright (C) 2010-2014 QuakeSpasm developers

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifndef _QUAKE_SBAR_H
#define _QUAKE_SBAR_H

// the status bar is only redrawn if something has changed, but if anything
// does, the entire thing will be redrawn for the next vid.numpages frames.

extern int sb_lines; // scan lines to draw

void Sbar_Init (void);
void Sbar_LoadPics (void);

void Sbar_Draw (cb_context_t *cbx);
// called every frame by screen

void Sbar_IntermissionOverlay (cb_context_t *cbx);
// called each frame after the level has been completed

void Sbar_FinaleOverlay (cb_context_t *cbx);

#endif /* _QUAKE_SBAR_H */
