/* Ogg/Vorbis streaming music support. */

#if !defined(_SND_VORBIS_H_)
#define _SND_VORBIS_H_ 1

#if defined(USE_CODEC_VORBIS)

extern snd_codec_t vorbis_codec;

#endif /* USE_CODEC_VORBIS */

#endif /* ! _SND_VORBIS_H_ */
