;
; Definition file of vulkan-1.dll
; Automatic generated by gendef
; written by Kai Tietz 2008
;
; generated from VulkanRT-1.2.176.1-Components.zip:x86/vulkan-1.dll
; to build an import library:
; i686-w64-mingw32-dlltool --as-flags=--32 -m i386 -k --as=i686-w64-mingw32-as --output-lib libvulkan-1.dll.a --input-def vulkan-1.def
;
LIBRARY "vulkan-1.dll"
EXPORTS
vkAcquireNextImage2KHR@12
vkAcquireNextImageKHR@40
vkAllocateCommandBuffers@12
vkAllocateDescriptorSets@12
vkAllocateMemory@16
vkBeginCommandBuffer@8
vkBindBufferMemory2@12
vkBindBufferMemory@28
vkBindImageMemory2@12
vkBindImageMemory@28
vkCmdBeginQuery@20
vkCmdBeginRenderPass2@12
vkCmdBeginRenderPass@12
vkCmdBindDescriptorSets@36
vkCmdBindIndexBuffer@24
vkCmdBindPipeline@16
vkCmdBindVertexBuffers@20
vkCmdBlitImage@40
vkCmdClearAttachments@20
vkCmdClearColorImage@28
vkCmdClearDepthStencilImage@28
vkCmdCopyBuffer@28
vkCmdCopyBufferToImage@32
vkCmdCopyImage@36
vkCmdCopyImageToBuffer@32
vkCmdCopyQueryPoolResults@48
vkCmdDispatch@16
vkCmdDispatchBase@28
vkCmdDispatchIndirect@20
vkCmdDraw@20
vkCmdDrawIndexed@24
vkCmdDrawIndexedIndirect@28
vkCmdDrawIndexedIndirectCount@44
vkCmdDrawIndirect@28
vkCmdDrawIndirectCount@44
vkCmdEndQuery@16
vkCmdEndRenderPass2@8
vkCmdEndRenderPass@4
vkCmdExecuteCommands@12
vkCmdFillBuffer@32
vkCmdNextSubpass2@12
vkCmdNextSubpass@8
vkCmdPipelineBarrier@40
vkCmdPushConstants@28
vkCmdResetEvent@16
vkCmdResetQueryPool@20
vkCmdResolveImage@36
vkCmdSetBlendConstants@8
vkCmdSetDepthBias@16
vkCmdSetDepthBounds@12
vkCmdSetDeviceMask@8
vkCmdSetEvent@16
vkCmdSetLineWidth@8
vkCmdSetScissor@16
vkCmdSetStencilCompareMask@12
vkCmdSetStencilReference@12
vkCmdSetStencilWriteMask@12
vkCmdSetViewport@16
vkCmdUpdateBuffer@32
vkCmdWaitEvents@44
vkCmdWriteTimestamp@20
vkCreateBuffer@16
vkCreateBufferView@16
vkCreateCommandPool@16
vkCreateComputePipelines@28
vkCreateDescriptorPool@16
vkCreateDescriptorSetLayout@16
vkCreateDescriptorUpdateTemplate@16
vkCreateDevice@16
vkCreateDisplayModeKHR@24
vkCreateDisplayPlaneSurfaceKHR@16
vkCreateEvent@16
vkCreateFence@16
vkCreateFramebuffer@16
vkCreateGraphicsPipelines@28
vkCreateImage@16
vkCreateImageView@16
vkCreateInstance@12
vkCreatePipelineCache@16
vkCreatePipelineLayout@16
vkCreateQueryPool@16
vkCreateRenderPass2@16
vkCreateRenderPass@16
vkCreateSampler@16
vkCreateSamplerYcbcrConversion@16
vkCreateSemaphore@16
vkCreateShaderModule@16
vkCreateSharedSwapchainsKHR@20
vkCreateSwapchainKHR@16
vkCreateWin32SurfaceKHR@16
vkDestroyBuffer@16
vkDestroyBufferView@16
vkDestroyCommandPool@16
vkDestroyDescriptorPool@16
vkDestroyDescriptorSetLayout@16
vkDestroyDescriptorUpdateTemplate@16
vkDestroyDevice@8
vkDestroyEvent@16
vkDestroyFence@16
vkDestroyFramebuffer@16
vkDestroyImage@16
vkDestroyImageView@16
vkDestroyInstance@8
vkDestroyPipeline@16
vkDestroyPipelineCache@16
vkDestroyPipelineLayout@16
vkDestroyQueryPool@16
vkDestroyRenderPass@16
vkDestroySampler@16
vkDestroySamplerYcbcrConversion@16
vkDestroySemaphore@16
vkDestroyShaderModule@16
vkDestroySurfaceKHR@16
vkDestroySwapchainKHR@16
vkDeviceWaitIdle@4
vkEndCommandBuffer@4
vkEnumerateDeviceExtensionProperties@16
vkEnumerateDeviceLayerProperties@12
vkEnumerateInstanceExtensionProperties@12
vkEnumerateInstanceLayerProperties@8
vkEnumerateInstanceVersion@4
vkEnumeratePhysicalDeviceGroups@12
vkEnumeratePhysicalDevices@12
vkFlushMappedMemoryRanges@12
vkFreeCommandBuffers@20
vkFreeDescriptorSets@20
vkFreeMemory@16
vkGetBufferDeviceAddress@8
vkGetBufferMemoryRequirements2@12
vkGetBufferMemoryRequirements@16
vkGetBufferOpaqueCaptureAddress@8
vkGetDescriptorSetLayoutSupport@12
vkGetDeviceGroupPeerMemoryFeatures@20
vkGetDeviceGroupPresentCapabilitiesKHR@8
vkGetDeviceGroupSurfacePresentModesKHR@16
vkGetDeviceMemoryCommitment@16
vkGetDeviceMemoryOpaqueCaptureAddress@8
vkGetDeviceProcAddr@8
vkGetDeviceQueue2@12
vkGetDeviceQueue@16
vkGetDisplayModeProperties2KHR@20
vkGetDisplayModePropertiesKHR@20
vkGetDisplayPlaneCapabilities2KHR@12
vkGetDisplayPlaneCapabilitiesKHR@20
vkGetDisplayPlaneSupportedDisplaysKHR@16
vkGetEventStatus@12
vkGetFenceStatus@12
vkGetImageMemoryRequirements2@12
vkGetImageMemoryRequirements@16
vkGetImageSparseMemoryRequirements2@16
vkGetImageSparseMemoryRequirements@20
vkGetImageSubresourceLayout@20
vkGetInstanceProcAddr@8
vkGetPhysicalDeviceDisplayPlaneProperties2KHR@12
vkGetPhysicalDeviceDisplayPlanePropertiesKHR@12
vkGetPhysicalDeviceDisplayProperties2KHR@12
vkGetPhysicalDeviceDisplayPropertiesKHR@12
vkGetPhysicalDeviceExternalBufferProperties@12
vkGetPhysicalDeviceExternalFenceProperties@12
vkGetPhysicalDeviceExternalSemaphoreProperties@12
vkGetPhysicalDeviceFeatures2@8
vkGetPhysicalDeviceFeatures@8
vkGetPhysicalDeviceFormatProperties2@12
vkGetPhysicalDeviceFormatProperties@12
vkGetPhysicalDeviceImageFormatProperties2@12
vkGetPhysicalDeviceImageFormatProperties@28
vkGetPhysicalDeviceMemoryProperties2@8
vkGetPhysicalDeviceMemoryProperties@8
vkGetPhysicalDevicePresentRectanglesKHR@20
vkGetPhysicalDeviceProperties2@8
vkGetPhysicalDeviceProperties@8
vkGetPhysicalDeviceQueueFamilyProperties2@12
vkGetPhysicalDeviceQueueFamilyProperties@12
vkGetPhysicalDeviceSparseImageFormatProperties2@16
vkGetPhysicalDeviceSparseImageFormatProperties@32
vkGetPhysicalDeviceSurfaceCapabilities2KHR@12
vkGetPhysicalDeviceSurfaceCapabilitiesKHR@16
vkGetPhysicalDeviceSurfaceFormats2KHR@16
vkGetPhysicalDeviceSurfaceFormatsKHR@20
vkGetPhysicalDeviceSurfacePresentModesKHR@20
vkGetPhysicalDeviceSurfaceSupportKHR@20
vkGetPhysicalDeviceWin32PresentationSupportKHR@8
vkGetPipelineCacheData@20
vkGetQueryPoolResults@40
vkGetRenderAreaGranularity@16
vkGetSemaphoreCounterValue@16
vkGetSwapchainImagesKHR@20
vkInvalidateMappedMemoryRanges@12
vkMapMemory@36
vkMergePipelineCaches@20
vkQueueBindSparse@20
vkQueuePresentKHR@8
vkQueueSubmit@20
vkQueueWaitIdle@4
vkResetCommandBuffer@8
vkResetCommandPool@16
vkResetDescriptorPool@16
vkResetEvent@12
vkResetFences@12
vkResetQueryPool@20
vkSetEvent@12
vkSignalSemaphore@8
vkTrimCommandPool@16
vkUnmapMemory@12
vkUpdateDescriptorSetWithTemplate@24
vkUpdateDescriptorSets@20
vkWaitForFences@24
vkWaitSemaphores@16
