/* module tracker decoding support using libmodplug */
#if !defined(_SND_MODPLUG_H_)
#define _SND_MODPLUG_H_

#if defined(USE_CODEC_MODPLUG)

extern snd_codec_t modplug_codec;

#endif /* USE_CODEC_MODPLUG */

#endif /* ! _SND_MODPLUG_H_ */
