/* module tracker decoding support using libmikmod */

#if !defined(_SND_MIKMOD_H_)
#define _SND_MIKMOD_H_

#if defined(USE_CODEC_MIKMOD)

extern snd_codec_t mikmod_codec;

#endif /* USE_CODEC_MIKMOD */

#endif /* ! _SND_MIKMOD_H_ */
