/* fLaC streaming music support. */

#if !defined(_SND_FLAC_H_)
#define _SND_FLAC_H_ 1

#if defined(USE_CODEC_FLAC)

extern snd_codec_t flac_codec;

#endif /* USE_CODEC_FLAC */

#endif /* ! _SND_FLAC_H_ */
